X_TARGET = 'target'     # a a target variable
X_META = 'meta'         # meta data (not to use as training feature)
X_KEY = 'key'           # a sample key (should be unique)
X_FEATURE = 'feature'   # a feature (can be used for training)

X_BINARY = 'binary'                # a binary feature
X_CATEGORICAL = 'categorical'      # a categorical feature
X_ORDINAL = 'ordinal'
X_INTERVAL = 'interval'
X_RATIO = 'ratio'

